from unittest import TestCase

from pyinfra.api.util import (
    format_exception,
    get_caller_frameinfo,
    try_int,
    underscore,
)


class TestApiUtil(TestCase):
    def test_try_int_number(self):
        assert try_int('1') == 1

    def test_try_int_fail(self):
        assert try_int('hello') == 'hello'

    def test_get_caller_frameinfo(self):
        def _get_caller_frameinfo():
            return get_caller_frameinfo()

        frameinfo = _get_caller_frameinfo()
        assert frameinfo.lineno == 22  # called by the line above

    def test_underscore(self):
        for camel_string, snake_string in (
            ('ThisIsHowYouDoIt', 'this_is_how_you_do_it'),
            ('This What_nope', 'this _what_nope'),
            ('myNameIs', 'my_name_is'),
        ):
            assert underscore(camel_string) == snake_string

    def test_format_exception(self):
        exception = Exception('I am a message', 1)
        assert format_exception(exception) == "Exception('I am a message', 1)"



import json
import os

from datetime import datetime
from io import open
from os import listdir, path

import six
from mock import patch

from pyinfra.api import Config, Inventory


def get_command_string(command):
    value = command.get_raw_value()
    masked_value = command.get_masked_value()
    if value == masked_value:
        return value
    else:
        return [value, masked_value]


def make_inventory(hosts=('somehost', 'anotherhost'), **kwargs):
    return Inventory(
        (hosts, {}),
        test_group=([
            'somehost',
        ], {
            'group_data': 'hello world',
        }),
        ssh_user='vagrant',
        **kwargs
    )


class FakeState(object):
    active = True
    deploy_dir = '/'
    in_op = True
    in_deploy = True
    pipelining = False
    deploy_name = None
    deploy_kwargs = None

    def __init__(self):
        self.inventory = Inventory(([], {}))
        self.config = Config()

    def get_temp_filename(*args):
        return '_tempfile_'

    def add_will_add_user(self, username):
        pass


def parse_fact(fact):
    '''
    Convert JSON types to more complex Python types because JSON is lacking.
    '''

    # Handle datetimes
    if isinstance(fact, six.string_types) and fact.startswith('datetime:'):
        return datetime.strptime(fact[9:], '%Y-%m-%dT%H:%M:%S')

    elif isinstance(fact, list):
        # Handle sets
        if len(fact) > 1 and fact[0] == '_set':
            return set(parse_fact(value) for value in fact[1:])

        return [parse_fact(value) for value in fact]

    elif isinstance(fact, dict):
        return {
            key: parse_fact(value)
            for key, value in six.iteritems(fact)
        }

    return fact


class FakeFact(object):
    def __init__(self, data):
        self.data = parse_fact(data)

    def __iter__(self):
        return iter(self.data)

    def __getattr__(self, key):
        return getattr(self.data, key)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __contains__(self, key):
        return key in self.data

    def __call__(self, *args, **kwargs):
        item = self.data

        for arg in args:
            if arg is None:
                continue

            # Support for non-JSON-able fact arguments by  turning them into JSON!
            if isinstance(arg, list):
                arg = json.dumps(arg)

            item = item[arg]

        return item

    def __str__(self):
        return self.data

    def __unicode__(self):
        return self.data

    def __eq__(self, other_thing):
        return self.data == other_thing

    def __ne__(self, other_thing):
        return self.data != other_thing

    def get(self, key, default=None):
        if key in self.data:
            return self.data[key]

        return default


class FakeFacts(object):
    def __init__(self, facts):
        self.facts = facts

    def __getattr__(self, key):
        if self.facts[key] is None:
            return None
        else:
            return FakeFact(self.facts[key])

    def _create(self, key, data=None, args=None):
        pass

    def _delete(self, key, args=None):
        pass


class FakeHost(object):
    noop_description = None

    def __init__(self, name, facts, data):
        self.name = name
        self.fact = FakeFacts(facts)
        self.data = data

    @property
    def print_prefix(self):
        return ''

    def noop(self, description):
        self.noop_description = description


class FakeFile(object):
    _read = False
    _data = None

    def __init__(self, name, data=None):
        self._name = name
        self._data = data

    def read(self, *args, **kwargs):
        if self._read is False:
            self._read = True

            if self._data:
                return self._data
            else:
                return '_test_data_'

        return ''

    def readlines(self, *args, **kwargs):
        if self._read is False:
            self._read = True

            if self._data:
                return self._data.split()
            else:
                return ['_test_data_']

            return []

    def seek(self, *args, **kwargs):
        pass

    def close(self, *args, **kwargs):
        pass

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        pass


class patch_files(object):
    def __init__(self, patch_files, patch_directories):
        files = []
        files_data = {}

        for filename_data in patch_files:
            if isinstance(filename_data, list):
                filename, data = filename_data
            else:
                filename = filename_data
                data = None

            if not filename.startswith(os.sep):
                filename = '{0}{1}'.format(FakeState.deploy_dir, filename)

            files.append(filename)

            if data:
                files_data[filename] = data

        self.files = files
        self.files_data = files_data

        self.directories = patch_directories

    def __enter__(self):
        self.patches = [
            patch('pyinfra.operations.files.os_path.exists', self.exists),
            patch('pyinfra.operations.files.os_path.isfile', self.isfile),
            patch('pyinfra.operations.files.os_path.isdir', self.isdir),
            patch('pyinfra.operations.files.walk', self.walk),
            patch('pyinfra.operations.files.makedirs', lambda path: True),
            # Builtin patches
            patch('pyinfra.operations.files.open', self.get_file, create=True),
            patch('pyinfra.operations.server.open', self.get_file, create=True),
            patch('pyinfra.api.util.open', self.get_file, create=True),
        ]

        for patched in self.patches:
            patched.start()

    def __exit__(self, type_, value, traceback):
        for patched in self.patches:
            patched.stop()

    def get_file(self, filename, *args):
        if filename in self.files:
            return FakeFile(filename, self.files_data.get(filename))

        raise IOError('Missing FakeFile: {0}'.format(filename))

    def exists(self, filename, *args):
        return filename in self.files or filename in self.directories

    def isfile(self, filename, *args):
        return filename in self.files

    def isdir(self, dirname, *args):
        return dirname in self.directories

    def walk(self, dirname):
        if dirname not in self.directories:
            return

        for dirname, filenames in sorted(self.directories[dirname].items()):
            yield dirname, None, filenames


def create_host(name=None, facts=None, data=None):
    '''
    Creates a FakeHost object with attached fact data.
    '''

    real_facts = {}
    facts = facts or {}

    for name, fact_data in six.iteritems(facts):
        real_facts[name] = fact_data

    return FakeHost(name, facts=real_facts, data=data)


class JsonTest(type):
    def __new__(cls, name, bases, attrs):
        # Get the JSON files
        files = listdir(attrs['jsontest_files'])
        files = [f for f in files if f.endswith('.json')]

        test_prefix = attrs.get('jsontest_prefix', 'test_')

        def gen_test(test_name, filename):
            def test(self):
                test_data = json.loads(open(
                    path.join(attrs['jsontest_files'], filename),
                    encoding='utf-8',
                ).read())
                self.jsontest_function(test_name, test_data)

            return test

        # Loop them and create class methods to call the jsontest_function
        for filename in files:
            test_name = filename[:-5]

            # Attach the method
            method_name = '{0}{1}'.format(test_prefix, test_name)
            attrs[method_name] = gen_test(test_name, filename)

        return type.__new__(cls, name, bases, attrs)



from pyinfra.api import FactBase

from .util.packaging import parse_packages

BREW_REGEX = r'^([^\s]+)\s([0-9\._+a-z\-]+)'


class BrewPackages(FactBase):
    '''
    Returns a dict of installed brew packages:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    command = 'brew list --versions'
    requires_command = 'brew'

    default = dict

    def process(self, output):
        return parse_packages(BREW_REGEX, output)


class BrewCasks(BrewPackages):
    '''
    Returns a dict of installed brew casks:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    command = 'brew cask list --versions'
    requires_command = 'brew'


class BrewTaps(FactBase):
    '''
    Returns a list of brew taps.
    '''

    command = 'brew tap'
    requires_command = 'brew'

    default = list

    def process(self, output):
        return output



from __future__ import unicode_literals

from pyinfra.api import FactBase

from .util.packaging import parse_packages

PIP_REGEX = r'^([a-zA-Z0-9_\-\+\.]+)==([0-9\.]+[a-z0-9\-]*)$'


class PipPackages(FactBase):
    '''
    Returns a dict of installed pip packages:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    default = dict
    pip_command = 'pip'

    def requires_command(self, pip=None):
        return pip or self.pip_command

    def command(self, pip=None):
        pip = pip or self.pip_command
        return '{0} freeze --all'.format(pip)

    def process(self, output):
        return parse_packages(PIP_REGEX, output)


class Pip3Packages(PipPackages):
    pip_command = 'pip3'



from __future__ import unicode_literals

from pyinfra.api import operation

from . import files
from .util.packaging import ensure_packages, ensure_rpm, ensure_yum_repo
from .yum import key as yum_key

key = yum_key


@operation
def repo(
    src,
    baseurl=None,
    present=True,
    description=None,
    enabled=True,
    gpgcheck=True,
    gpgkey=None,
    type='rpm-md',
    state=None,
    host=None,
):
    '''
    Add/remove/update zypper repositories.

    + src: URL or name for the ``.repo``   file
    + baseurl: the baseurl of the repo (if ``name`` is not a URL)
    + present: whether the ``.repo`` file should be present
    + description: optional verbose description
    + enabled: whether this repo is enabled
    + gpgcheck: whether set ``gpgcheck=1``
    + gpgkey: the URL to the gpg key for this repo
    + type: the type field this repo (defaults to ``rpm-md``)

    ``Baseurl``/``description``/``gpgcheck``/``gpgkey``:
        These are only valid when ``name`` is a filename (ie not a URL). This is
        for manual construction of repository files. Use a URL to download and
        install remote repository files.

    Examples:

    .. code:: python

        # Download a repository file
        zypper.repo(
            name='Install container virtualization repo via URL',
            src='https://download.opensuse.org/repositories/Virtualization:containers/openSUSE_Tumbleweed/Virtualization:containers.repo',
        )

        # Create the repository file from baseurl/etc
        zypper.repo(
            name='Install container virtualization repo',
            src=='Virtualization:containers (openSUSE_Tumbleweed)',
            baseurl='https://download.opensuse.org/repositories/Virtualization:/containers/openSUSE_Tumbleweed/',
        )
    '''

    yield ensure_yum_repo(
        state,
        host,
        files,
        src,
        baseurl,
        present,
        description,
        enabled,
        gpgcheck,
        gpgkey,
        type_=type,
        repo_directory='/etc/zypp/repos.d/',
    )


@operation
def rpm(src, present=True, state=None, host=None):
    # NOTE: if updating this docstring also update `dnf.rpm`
    '''
    Add/remove ``.rpm`` file packages.

    + src: filename or URL of the ``.rpm`` package
    + present: whether ore not the package should exist on the system

    URL sources with ``present=False``:
        If the ``.rpm`` file isn't downloaded, pyinfra can't remove any existing
        package as the file won't exist until mid-deploy.

    Example:

    .. code:: python

        zypper.rpm(
           name='Install task from rpm',
           src='https://github.com/go-task/task/releases/download/v2.8.1/task_linux_amd64.rpm'
        )
    '''

    yield ensure_rpm(state, host, files, src, present, 'zypper --non-interactive')


@operation
def update(state=None, host=None):
    '''
    Updates all zypper packages.
    '''

    yield 'zypper update -y'


_update = update  # noqa: E305 (for use below where update is a kwarg)


@operation
def packages(
    packages=None,
    present=True,
    latest=False,
    update=False,
    clean=False,
    extra_global_install_args=None,
    extra_install_args=None,
    extra_global_uninstall_args=None,
    extra_uninstall_args=None,
    state=None,
    host=None,
):
    '''
    Install/remove/update zypper packages & updates.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``zypper update`` before installing packages
    + clean: run ``zypper clean --all`` before installing packages
    + extra_global_install_args: additional global arguments to the zypper install command
    + extra_install_args: additional arguments to the zypper install command
    + extra_global_uninstall_args: additional global arguments to the zypper uninstall command
    + extra_uninstall_args: additional arguments to the zypper uninstall command

    Versions:
        Package versions can be pinned like zypper: ``<pkg>=<version>``

    Examples:

    .. code:: python

        # Update package list and install packages
        zypper.packages(
            name='Install Vim and Vim enhanced',
            packages=['vim-enhanced', 'vim'],
            update=True,
        )

        # Install the latest versions of packages (always check)
        zypper.packages(
            name='Install latest Vim',
            packages=['vim'],
            latest=True,
        )
    '''

    if clean:
        yield 'zypper clean --all'

    if update:
        yield _update(state=state, host=host)

    install_command = ['zypper', '--non-interactive', 'install', '-y']

    if extra_install_args:
        install_command.append(extra_install_args)

    if extra_global_install_args:
        install_command.insert(1, extra_global_install_args)

    uninstall_command = ['zypper', '--non-interactive', 'remove', '-y']

    if extra_uninstall_args:
        uninstall_command.append(extra_uninstall_args)

    if extra_global_uninstall_args:
        uninstall_command.insert(1, extra_global_uninstall_args)

    upgrade_command = 'zypper update -y'

    yield ensure_packages(
        host,
        packages,
        host.fact.rpm_packages,
        present,
        install_command=' '.join(install_command),
        uninstall_command=' '.join(uninstall_command),
        upgrade_command=upgrade_command,
        version_join='=',
        latest=latest,
    )



def handle_service_control(
    host,
    name, statuses, formatter, running, restarted, reloaded, command,
    status_argument='status',
):
    status = statuses.get(name, None)

    # If we don't know the status, we need to check if it's up before starting
    # and/or restarting/reloading
    if status is None:
        yield '''
            # If the service is running
            if {status_command}; then
                {stop_command}
                {restart_command}
                {reload_command}

            # If the service is not running, we just start it (no re[start|load])
            else
                {start_command}
            fi
        '''.format(
            status_command=formatter.format(name, status_argument),
            start_command=(
                formatter.format(name, 'start')
                if running is True else 'true'
            ),
            stop_command=(
                formatter.format(name, 'stop')
                if running is False else 'true'
            ),
            restart_command=(
                formatter.format(name, 'restart')
                if restarted else 'true'
            ),
            reload_command=(
                formatter.format(name, 'reload')
                if reloaded else 'true'
            ),
        )
        statuses[name] = running

    else:
        # Need down but running
        if running is False:
            if status:
                yield formatter.format(name, 'stop')
                statuses[name] = False
            else:
                host.noop('service {0} is stopped'.format(name))

        # Need running but down
        if running is True:
            if not status:
                yield formatter.format(name, 'start')
                statuses[name] = True
            else:
                host.noop('service {0} is running'.format(name))

        # Only restart if the service is already running
        if restarted and status:
            yield formatter.format(name, 'restart')

        # Only reload if the service is already reloaded
        if reloaded and status:
            yield formatter.format(name, 'reload')

    # Always execute arbitrary commands as these may or may not rely on the service
    # being up or down
    if command:
        yield formatter.format(name, command)



import re

from pyinfra.api import FactBase


class InitdStatus(FactBase):
    '''
    Low level check for every /etc/init.d/* script. Unfortunately many of these
    mishehave and return exit status 0 while also displaying the help info/not
    offering status support.

    Returns a dict of name -> status.

    Expected codes found at:
        http://refspecs.linuxbase.org/LSB_3.1.0/LSB-Core-generic/LSB-Core-generic/iniscrptact.html
    '''

    command = '''
        for SERVICE in `ls /etc/init.d/`; do
            _=`cat /etc/init.d/$SERVICE | grep "### BEGIN INIT INFO"`

            if [ "$?" = "0" ]; then
                STATUS=`/etc/init.d/$SERVICE status`
                echo "$SERVICE=$?"
            fi
        done
    '''

    regex = r'([a-zA-Z0-9\-]+)=([0-9]+)'
    default = dict

    def process(self, output):
        services = {}

        for line in output:
            matches = re.match(self.regex, line)
            if matches:
                status = int(matches.group(2))

                # Exit code 0 = OK/running
                if status == 0:
                    status = True

                # Exit codes 1-3 = DOWN/not running
                elif status < 4:
                    status = False

                # Exit codes 4+ = unknown
                else:
                    status = None

                services[matches.group(1)] = status

        return services



import json

from unittest import TestCase

from mock import mock_open, patch

from pyinfra.api.connectors.vagrant import get_vagrant_options, make_names_data
from pyinfra.api.exceptions import InventoryError

FAKE_VAGRANT_OPTIONS = {
    'groups': {
        'ubuntu16': ['mygroup'],
    },
    'data': {
        'centos7': {
            'somedata': 'somevalue',
        },
    },
}
FAKE_VAGRANT_OPTIONS_DATA = json.dumps(FAKE_VAGRANT_OPTIONS)


def fake_vagrant_shell(command, splitlines=None):
    if command == 'vagrant status --machine-readable':
        return [
            '1574707743,ubuntu16,provider-name,virtualbox',
            '1574707743,ubuntu16,state,running',
            '1574707743,ubuntu18,state,not_created',
            '1574707743,centos7,state,running',
            '1574707743,centos6,state,running',
        ]
    elif command == 'vagrant ssh-config ubuntu16':
        return [
            'ExtraKey logme!',
            'Host ubuntu16',
            'HostName 127.0.0.1',
            'User vagrant',
            'Port 2222',
            'PasswordAuthentication no',
            'IdentityFile path/to/key',
            '',
        ]
    elif command == 'vagrant ssh-config centos7':
        return [
            '',
            'Host centos7',
            'HostName 127.0.0.1',
            'User vagrant',
            'Port 2200',
            'PasswordAuthentication no',
            'IdentityFile path/to/key',
        ]
    elif command == 'vagrant ssh-config centos6':
        return [
            'Host centos6',
            'HostName 127.0.0.1',
        ]

    return []


@patch('pyinfra.api.connectors.vagrant.local.shell', fake_vagrant_shell)
class TestVagrantConnector(TestCase):
    def tearDown(self):
        get_vagrant_options.cache = {}

    @patch(
        'pyinfra.api.connectors.vagrant.open',
        mock_open(read_data=FAKE_VAGRANT_OPTIONS_DATA),
        create=True,
    )
    @patch('pyinfra.api.connectors.vagrant.path.exists', lambda path: True)
    def test_make_names_data_with_options(self):
        data = make_names_data()

        assert data == [
            (
                '@vagrant/ubuntu16',
                {
                    'ssh_port': '2222',
                    'ssh_user': 'vagrant',
                    'ssh_hostname': '127.0.0.1',
                    'ssh_key': 'path/to/key',
                },
                ['mygroup', '@vagrant'],
            ), (
                '@vagrant/centos7',
                {
                    'ssh_port': '2200',
                    'ssh_user': 'vagrant',
                    'ssh_hostname': '127.0.0.1',
                    'ssh_key': 'path/to/key',
                    'somedata': 'somevalue',
                },
                ['@vagrant'],
            ), (
                '@vagrant/centos6',
                {
                    'ssh_hostname': '127.0.0.1',
                },
                ['@vagrant'],
            ),
        ]

    def test_make_names_data_with_limit(self):
        data = make_names_data(limit=('ubuntu16',))

        assert data == [
            (
                '@vagrant/ubuntu16',
                {
                    'ssh_port': '2222',
                    'ssh_user': 'vagrant',
                    'ssh_hostname': '127.0.0.1',
                    'ssh_key': 'path/to/key',
                },
                ['@vagrant'],
            ),
        ]

    def test_make_names_data_no_matches(self):
        with self.assertRaises(InventoryError):
            make_names_data(limit='nope')



from pyinfra.api import (
    Config,
    State,
    StringCommand,
)
from pyinfra.api.connect import connect_all, disconnect_all
from pyinfra.api.deploy import add_deploy, deploy
from pyinfra.api.operations import run_ops
from pyinfra.operations import server

from ..paramiko_util import PatchSSHTestCase
from ..util import make_inventory


class TestDeploysApi(PatchSSHTestCase):
    def test_deploy(self):
        inventory = make_inventory()
        somehost = inventory.get_host('somehost')
        anotherhost = inventory.get_host('anotherhost')

        state = State(inventory, Config())

        # Enable printing on this test to catch any exceptions in the formatting
        state.print_output = True
        state.print_input = True
        state.print_fact_info = True
        state.print_noop_info = True

        connect_all(state)

        @deploy
        def test_deploy(state=None, host=None):
            server.shell(
                commands=['echo first command'],
                state=state,
                host=host,
            )
            server.shell(
                commands=['echo second command'],
                state=state,
                host=host,
            )

        add_deploy(state, test_deploy)

        op_order = state.get_op_order()

        # Ensure we have an op
        assert len(op_order) == 2

        first_op_hash = op_order[0]
        assert state.op_meta[first_op_hash]['names'] == {'test_deploy | Server/Shell'}
        assert state.ops[somehost][first_op_hash]['commands'] == [
            StringCommand('echo first command'),
        ]
        assert state.ops[anotherhost][first_op_hash]['commands'] == [
            StringCommand('echo first command'),
        ]

        second_op_hash = op_order[1]
        assert state.op_meta[second_op_hash]['names'] == {'test_deploy | Server/Shell'}
        assert state.ops[somehost][second_op_hash]['commands'] == [
            StringCommand('echo second command'),
        ]
        assert state.ops[anotherhost][second_op_hash]['commands'] == [
            StringCommand('echo second command'),
        ]

        # Ensure run ops works
        run_ops(state)

        # Ensure ops completed OK
        assert state.results[somehost]['success_ops'] == 2
        assert state.results[somehost]['ops'] == 2
        assert state.results[anotherhost]['success_ops'] == 2
        assert state.results[anotherhost]['ops'] == 2

        # And w/o errors
        assert state.results[somehost]['error_ops'] == 0
        assert state.results[anotherhost]['error_ops'] == 0

        # And with the different modes
        run_ops(state, serial=True)
        run_ops(state, no_wait=True)

        disconnect_all(state)

    def test_nested_deploy(self):
        inventory = make_inventory()
        somehost = inventory.get_host('somehost')

        state = State(inventory, Config())

        # Enable printing on this test to catch any exceptions in the formatting
        state.print_output = True
        state.print_input = True
        state.print_fact_info = True
        state.print_noop_info = True

        connect_all(state)

        @deploy
        def test_nested_deploy(state=None, host=None):
            server.shell(
                commands=['echo nested command'],
                state=state,
                host=host,
            )

        @deploy
        def test_deploy(state=None, host=None):
            server.shell(
                commands=['echo first command'],
                state=state,
                host=host,
            )
            test_nested_deploy(
                state=state,
                host=host,
            )
            server.shell(
                commands=['echo second command'],
                state=state,
                host=host,
            )

        add_deploy(state, test_deploy)

        op_order = state.get_op_order()

        # Ensure we have an op
        assert len(op_order) == 3

        first_op_hash = op_order[0]
        assert state.op_meta[first_op_hash]['names'] == {'test_deploy | Server/Shell'}
        assert state.ops[somehost][first_op_hash]['commands'] == [
            StringCommand('echo first command'),
        ]

        second_op_hash = op_order[1]
        assert state.op_meta[second_op_hash]['names'] == {
            'test_deploy | test_nested_deploy | Server/Shell',
        }
        assert state.ops[somehost][second_op_hash]['commands'] == [
            StringCommand('echo nested command'),
        ]

        third_op_hash = op_order[2]
        assert state.op_meta[third_op_hash]['names'] == {'test_deploy | Server/Shell'}
        assert state.ops[somehost][third_op_hash]['commands'] == [
            StringCommand('echo second command'),
        ]
